/***************************************************************************
  AS5600_interpolation.ino
  Code by lingib https://www.instructables.com/
  Last update 26 June 2022

  ------
  ABOUT
  ------
  The AS5600 code is from https://curiousscientist.tech/blog/as5600-magnetic-encoder-a-practical-example
  The motor control code is a modified version of my https://www.instructables.com/4-Wire-Horizontal-Plotter/
  The interpolation code is mine.

  When testing I observed errors of up to 2 degrees.
  These errors are probably due to the magnet being tilted or misaligned or the presence of ferrous metal such has nuts and bolts
  I found that I was able to reduce these errors to less than 0.2 degrees using interpolation.

  ------
  NOTES
  ------
   - All motor step lines are on PORTB starting from pin 8
   - All motor direction lines are on PORTC staring from pin A0
   - All motors (that need to be stepped) are stepped simultaneously

  ----------
  COPYRIGHT
  ----------
  This code is free software: you can redistribute it and/or
  modify it under the terms of the GNU General Public License as published
  by the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This software is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License. If
  not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

// ----- Bit set/clear/check/toggle macros
#define SET(x,y) (x |=(1<<y))
#define CLR(x,y) (x &= (~(1<<y)))
#define CHK(x,y) (x & (1<<y))
#define TOG(x,y) (x^=(1<<y))

// ----- constants
#define PI 3.1415926535897932384626433832795
#define HALF_PI 1.5707963267948966192313216916398
#define TWO_PI 6.283185307179586476925286766559
#define DEG_TO_RAD 0.017453292519943295769236907684886
#define RAD_TO_DEG 57.295779513082320876798154814105
#define CW true
#define CCW false

// ----- Libraries
#include <Wire.h>                                       // This is for i2C

// ----- AS5600 Magnetic sensor
int magnetStatus = 0;                                   //value of the status register (MD, ML, MH)

int lowbyte;                                            //raw angle bits[7:0]
word highbyte;                                          //raw angle bits[11:8]
int rawAngle;                                           //final raw angle bits[11:0]
float degAngle;                                         //raw angle in degrees (360/4096 * [value between 0-4095])

/*
   Both of the following conditions are required:
    - The sampleInterval must divide evenly into 360 degrees
    - The sampleInterval/1.8 must be a whole number
*/
const int sampleInterval = 9;                           // degrees
const int samples = int(360 / sampleInterval + 2);
float calibrationTable[samples];                        // for recording magnet angles every "sampleInterval" degrees
float interpolatedAngle;

float numberOfTurns = 0;                                // number of turns
float startAngle = 0;                                   // starting angle
float taredAngle = 0;                                   // tared angle - based on the startup value
float totalAngle = 0;                                   // total absolute angular displacement
float previousTotalAngle = 0;                           // for the display printing

int quadrantNumber = 0;                                 // quadrant IDs
int previousQuadrantNumber = 0;                         // these are used for tracking the numberOfTurns

// ----- motor controller definitions
const byte DIR1 = 14;                                   // A0  (PORTC)
const byte STEP1 = 8;                                   // PB0 (PORTB)

#define STEPS_PER_MM  (200*16)/40                       // 200steps/rev; 16 x microstepping; GT2-20 pulley circumference
#define ROTATIONS 1                                     // numberof complete rotations
#define NUDGE STEPS_PER_MM*40*ROTATIONS;                // GT2-20 circumference is 40mm;
unsigned int PULSE_WIDTH = 2;                           // easydriver step pulse-width (uS)

unsigned long DELAY_MIN = 2500L;                        // minimum inter-step delay (uS) between motor steps (controls speed)
unsigned long DELAY1 = 0L;                              // inter-step delay for motor1 (uS)

unsigned long StepCounter = 0L;                         // motor steps
unsigned long numberSteps;                              // used for returning motor to home position

bool
DIRECTION1,                                             // current motor directions
STEP_M1 = false;                                        // motor step flags

// ----------------
// setup()
// ----------------
void setup()
{
  Serial.begin(115200);                                 // start serial
  Wire.begin();                                         // start i2C
  Wire.setClock(400000L);                               // fast clock

  // ----- flush the buffers
  Serial.flush();                                       // clear TX buffer
  while (Serial.available()) Serial.read();             // clear RX buffer

  // ----- initialise motor
  pinMode(DIR1, OUTPUT);                                // initialise motor1
  pinMode(STEP1, OUTPUT);
  digitalWrite(DIR1, CW);
  delayMicroseconds(PULSE_WIDTH);
  digitalWrite(STEP1, LOW);

  //checkMagnetPresence();                              // check the magnet (blocks until magnet is found)
  checkMagnetPresence();

  Serial.println("Calibrating");

  readRawAngle();                                       // make initial reading so the degAngle gets updated
  startAngle = degAngle;                                // update startAngle with degAngle - for taring

  /////////////////////////////////////////////////////////////////////////
  // ----- record magnet angle every "sampleInterval" degrees
  for (int i = 0; i < samples; i++) {
    readRawAngle();
    tareAngle();
    checkQuadrant();
    calibrationTable[i] = totalAngle;

    // ----- rotate sampleInterval degrees
    /*
        NEMA17 stepping motors are 1.8 degrees per step
        9 degrees requires 9/1.8 = 5 steps
        microstepping increases the number of steps to 5*16 = 80 steps
    */
    int motorSteps = int(sampleInterval * 16 / 1.8);
    for (int step = 0; step < motorSteps; step++)
    {
      step_motor(1, CCW);
      StepCounter++;
    }
  }

  ///////////////////////////////////////////////////////////////////////////
  // ------ display array contents
  for (int i = 0; i < samples; i++) {
    Serial.print(i * sampleInterval);
    Serial.print("\t");
    Serial.println(calibrationTable[i]);
  }
  Serial.println("");

  // ------ return to start position
  Serial.println("Returning to home position");
  numberSteps = StepCounter;
  for (int steps = 0; steps < numberSteps ; steps++) {
    step_motor(1, CW);
    StepCounter--;
  }
  readRawAngle();
  tareAngle();
  Serial.print("Home angle  ");
  Serial.println(taredAngle);
  Serial.print("Step Counter ");
  Serial.println(StepCounter);
  Serial.println("");


  ///////////////////////////////////////////////////////////////////////////////////
  Serial.println("Calibration check");
  Serial.println("");

  int stepsPerRev = 3200;                                // 200 motor-steps/rev * 16x microstepping
  int steps  = int((samples) * sampleInterval * 16 / 1.8);


  for (int i = 0; i < steps; i++) {                      //rotate to 112.5 degrees

    if (i % 50 == 0) {
      readRawAngle();
      tareAngle();
      interpolate();

      Serial.print("Step Counter ");
      Serial.print(StepCounter);
      Serial.print("\t");

      Serial.print("True angle  ");
      Serial.print((float)(360 * StepCounter) / 3200);
      Serial.print("\t");

      Serial.print("Tared angle  ");
      Serial.print(taredAngle);
      Serial.print("\t");

      Serial.print("Interpolated angle ");
      Serial.print(interpolatedAngle);
      Serial.print("\t");

      Serial.print("error ");
      Serial.print((float)(360 * StepCounter) / 3200 - interpolatedAngle);
      Serial.println("\t");
    }

    step_motor(1, CCW);
    StepCounter++;
  }

  ///////////////////////////////////////////////////////////////////////////////////////////////////
  // ------ return to start position
  Serial.println("Returning to home position");
  numberSteps = StepCounter;
  for (int steps = 0; steps < numberSteps ; steps++) {
    step_motor(1, CW);
    StepCounter--;
  }
  readRawAngle();
  tareAngle();
  Serial.print("Home angle  ");
  Serial.println(taredAngle);
  Serial.print("Step Counter ");
  Serial.println(StepCounter);
  Serial.println("");
}

// ----------------
// loop()
// ----------------
void loop()
{
  // the loop() has no code
}

// ==========================================================================
// interpolate()
// ==========================================================================
void interpolate()
{
  for (int index = 1; index < samples; index++) {
    if (calibrationTable[index] > taredAngle) {
      float difference1 = taredAngle - calibrationTable[index - 1];
      float difference2 = calibrationTable[index] - calibrationTable[index - 1];
      interpolatedAngle = (float)sampleInterval * ((index - 1) + difference1 / difference2);
      break;
    }
  }
}

// ==========================================================================
// display_angle()
// ==========================================================================
void display_angle() {
  readRawAngle();                                                 //ask the value from the sensor
  tareAngle();                                                    //tare the value
  checkQuadrant();                                                //check quadrant, check rotations, check absolute angular position
}

//====================================
// STEP MOTOR (step a single motor)
//====================================
void step_motor(int motor, bool dir) {

  // ----- locals
  byte patternC = PORTC;                                          // Direction lines
  enum {dir1, dir2, dir3, dir4};                                  // Define direction bit positions

  byte patternB = PORTB;                                          // STEP lines
  enum {step1, step2, step3, step4};                              // Define bit positions for PORTC

  switch (motor) {
    case 1:
      CLR(patternB, step1);
      CLR(patternB, step2);
      CLR(patternB, step3);
      CLR(patternB, step4);
      PORTB = patternB;

      (dir == CCW) ? CLR(patternC, dir1) : SET(patternC, dir1);   // Set motor direction
      PORTC = patternC;
      delayMicroseconds(PULSE_WIDTH);                             // Wait for direction line to stabilise

      SET(patternB, step1);                                       // Prepare step pulse
      PORTB = patternB;                                           // Step the motors
      delayMicroseconds(PULSE_WIDTH);
      CLR(patternB, step1);                                       // Generate trailing-edge
      PORTB = patternB;                                           // Step the motors
      delayMicroseconds(DELAY_MIN);
      break;
  }
}

// ----------------
// readRawAngle()
// ----------------
void readRawAngle()
{
  //----- read low-order bits 7:0
  Wire.beginTransmission(0x36);                         //connect to the sensor
  Wire.write(0x0D);                                     //figure 21 - register map: Raw angle (7:0)
  Wire.endTransmission();                               //end transmission
  Wire.requestFrom(0x36, 1);                            //request from the sensor
  while (Wire.available() == 0);                        //wait until it becomes available
  lowbyte = Wire.read();                                //Reading the data after the request

  // ----- read high-order bits 11:8
  Wire.beginTransmission(0x36);
  Wire.write(0x0C);                                     //figure 21 - register map: Raw angle (11:8)
  Wire.endTransmission();
  Wire.requestFrom(0x36, 1);
  while (Wire.available() == 0);
  highbyte = Wire.read();

  // ----- combine bytes
  highbyte = highbyte << 8;                             // shift highbyte to left
  rawAngle = highbyte | lowbyte;                        // combine bytes to get 12-bit value 11:0
  degAngle = rawAngle * 0.087890625;                    // 360/4096 = 0.087890625

  //Serial.print("Deg angle: ");
  //Serial.println(degAngle, 2);                          //absolute position of the encoder within the 0-360 circle
}

// ----------------
// tareAngle()
// ----------------
void tareAngle()
{
  // ----- recalculate angle
  taredAngle = degAngle - startAngle;             //this tares the position

  if (taredAngle < 0)                             //if the calculated angle is negative, we need to "normalize" it
  {
    taredAngle = taredAngle + 360;            //correction for negative numbers (i.e. -15 becomes +345)
  }
}

// ----------------
// checkQuadrant()
// ----------------
void checkQuadrant()
{
  /*
    //Quadrants:
    4  |  1
    ---|---
    3  |  2
  */

  // ----- Quadrant 1
  if (taredAngle >= 0 && taredAngle <= 90) {
    quadrantNumber = 1;
  }

  //Quadrant 2
  if (taredAngle > 90 && taredAngle <= 180) {
    quadrantNumber = 2;
  }

  //Quadrant 3
  if (taredAngle > 180 && taredAngle <= 270) {
    quadrantNumber = 3;
  }

  // ----- Quadrant 4
  if (taredAngle > 270 && taredAngle < 360) {
    quadrantNumber = 4;
  }

  if (quadrantNumber != previousQuadrantNumber)                   // if we changed quadrant
  {
    if (quadrantNumber == 1 && previousQuadrantNumber == 4) {
      numberOfTurns++;                                            // 4 --> 1 transition: CW rotation
    }

    if (quadrantNumber == 4 && previousQuadrantNumber == 1) {
      numberOfTurns--;                                            // 1 --> 4 transition: CCW rotation
    }

    previousQuadrantNumber = quadrantNumber;                      //update to the current quadrant
  }

  totalAngle = (numberOfTurns * 360) + taredAngle;            //number of turns (+/-) plus the actual angle within the 0-360 range
}

// ----------------------
// checkMagnetPresence()
// ----------------------
void checkMagnetPresence()
{
  //Status register output: 0 0 MD ML MH 0 0 0
  //MH: Too strong magnet - 100111 - DEC: 39
  //ML: Too weak magnet - 10111 - DEC: 23
  //MD: OK magnet - 110111 - DEC: 55

  // ---- Check MD status bit (magnet detected)
  while ((magnetStatus & B00100000) != 32)                      // locks MCU until magnet correctly positioned
  {
    magnetStatus = 0; //reset reading

    Wire.beginTransmission(0x36);                               //connect to the sensor
    Wire.write(0x0B);                                           //figure 21 - register map: Status: MD ML MH
    Wire.endTransmission();                                     //end transmission
    Wire.requestFrom(0x36, 1);                                  //request from the sensor
    while (Wire.available() == 0);                              //wait until it becomes available
    magnetStatus = Wire.read();                                 //Reading the data after the request
    Serial.print("Magnet Status  ");
    Serial.println(magnetStatus, BIN);
    Serial.println(" ");
  }
  delay(1000);
}
